import logging
from pathlib import Path

import microcore as mc
from microcore import ApiType, ui, utils
from git import Repo, GitCommandError
import typer

from ..core import get_base_branch
from ..utils import version, extract_gh_owner_repo
from ..cli_base import app
from ..gh_api import gh_api


@app.command(name="deploy", help="Deploy Gito workflows to GitHub Actions")
@app.command(name="init", hidden=True)
def deploy(
    api_type: ApiType = None,
    commit: bool = None,
    rewrite: bool = False,
    to_branch: str = typer.Option(
        default="gito_deploy",
        help="Branch name for new PR containing with Gito workflows commit"
    ),
    token: str = typer.Option(
        "", help="GitHub token (or set GITHUB_TOKEN env var)"
    ),
):
    repo = Repo(".")
    workflow_files = dict(
        code_review=Path(".github/workflows/gito-code-review.yml"),
        react_to_comments=Path(".github/workflows/gito-react-to-comments.yml")
    )
    for file in workflow_files.values():
        if file.exists():
            message = f"Gito workflow already exists at {utils.file_link(file)}."
            if rewrite:
                ui.warning(message)
            else:
                message += "\nUse --rewrite to overwrite it."
                ui.error(message)
                return False

    api_types = [ApiType.ANTHROPIC, ApiType.OPEN_AI, ApiType.GOOGLE_AI_STUDIO]
    default_models = {
        ApiType.ANTHROPIC: "claude-sonnet-4-20250514",
        ApiType.OPEN_AI: "gpt-4.1",
        ApiType.GOOGLE_AI_STUDIO: "gemini-2.5-pro",
    }
    secret_names = {
        ApiType.ANTHROPIC: "ANTHROPIC_API_KEY",
        ApiType.OPEN_AI: "OPENAI_API_KEY",
        ApiType.GOOGLE_AI_STUDIO: "GOOGLE_AI_API_KEY",
    }
    if not api_type:
        api_type = mc.ui.ask_choose(
            "Choose your LLM API type",
            api_types,
        )
    elif api_type not in api_types:
        mc.ui.error(f"Unsupported API type: {api_type}")
        return False
    major, minor, *_ = version().split(".")
    template_vars = dict(
        model=default_models[api_type],
        api_type=api_type,
        secret_name=secret_names[api_type],
        major=major,
        minor=minor,
        ApiType=ApiType,
        remove_indent=True,
    )
    gito_code_review_yml = mc.tpl(
        "github_workflows/gito-code-review.yml.j2",
        **template_vars
    )
    gito_react_to_comments_yml = mc.tpl(
        "github_workflows/gito-react-to-comments.yml.j2",
        **template_vars
    )

    workflow_files["code_review"].parent.mkdir(parents=True, exist_ok=True)
    workflow_files["code_review"].write_text(gito_code_review_yml)
    workflow_files["react_to_comments"].write_text(gito_react_to_comments_yml)
    print(
        mc.ui.green("Gito workflows have been created.\n")
        + f"  - {mc.utils.file_link(workflow_files['code_review'])}\n"
        + f"  - {mc.utils.file_link(workflow_files['react_to_comments'])}\n"
    )
    owner, repo_name = extract_gh_owner_repo(repo)
    if commit is True or commit is None and mc.ui.ask_yn(
        "Do you want to commit and push created GitHub workflows to a new branch?"
    ):
        repo.git.add([str(file) for file in workflow_files.values()])
        if not repo.active_branch.name.startswith(to_branch):
            repo.git.checkout("-b", to_branch)
        try:
            repo.git.commit("-m", "Deploy Gito workflows")
        except GitCommandError as e:
            if "nothing added" in str(e):
                ui.warning("Failed to commit changes: nothing was added")
            else:
                ui.error(f"Failed to commit changes: {e}")
                return False

        repo.git.push("origin", to_branch)
        print(f"Changes pushed to {to_branch} branch.")
        try:
            api = gh_api(repo=repo)
            base = get_base_branch(repo).split('/')[-1]
            logging.info(f"Creating PR {ui.green(to_branch)} -> {ui.yellow(base)}...")
            res = api.pulls.create(
                head=to_branch,
                base=base,
                title="Deploy Gito workflows",
            )
            print(f"Pull request #{res.number} created successfully:\n{res.html_url}")
        except Exception as e:
            mc.ui.error(f"Failed to create pull request automatically: {e}")
            print(
                f"Please create a PR from '{to_branch}' to your main branch and merge it:\n"
                f"https://github.com/{owner}/{repo_name}/compare/{to_branch}?expand=1"
            )
    else:
        print(
            "Now you can commit and push created GitHub workflows to your main repository branch.\n"
        )

    print(
        "(!IMPORTANT):\n"
        f"Add {mc.ui.cyan(secret_names[api_type])} with actual API_KEY "
        "to your repository secrets here:\n"
        f"https://github.com/{owner}/{repo_name}/settings/secrets/actions"
    )
    return True



import logging
import os

import git
from jira import JIRA, JIRAError

from gito.issue_trackers import IssueTrackerIssue, resolve_issue_key


def fetch_issue(issue_key, jira_url, username, api_token) -> IssueTrackerIssue | None:
    try:
        jira = JIRA(jira_url, basic_auth=(username, api_token))
        issue = jira.issue(issue_key)
        return IssueTrackerIssue(
            title=issue.fields.summary,
            description=issue.fields.description or "",
            url=f"{jira_url.rstrip('/')}/browse/{issue_key}"
        )
    except JIRAError as e:
        logging.error(
            f"Failed to fetch Jira issue {issue_key}: code {e.status_code} :: {e.text}"
        )
        return None
    except Exception as e:
        logging.error(f"Failed to fetch Jira issue {issue_key}: {e}")
        return None


def fetch_associated_issue(
    repo: git.Repo,
    jira_url=None,
    jira_username=None,
    jira_api_token=None,
    **kwargs
):
    """
    Pipeline step to fetch a Jira issue based on the current branch name.
    """
    jira_url = jira_url or os.getenv("JIRA_URL")
    jira_username = (
        jira_username
        or os.getenv("JIRA_USERNAME")
        or os.getenv("JIRA_USER")
        or os.getenv("JIRA_EMAIL")
    )
    jira_token = (
        jira_api_token
        or os.getenv("JIRA_API_TOKEN")
        or os.getenv("JIRA_API_KEY")
        or os.getenv("JIRA_TOKEN")
    )
    try:
        assert jira_url, "JIRA_URL is not set"
        assert jira_username, "JIRA_USERNAME is not set"
        assert jira_token, "JIRA_API_TOKEN is not set"
    except AssertionError as e:
        logging.error(f"Jira configuration error: {e}")
        return None
    issue_key = resolve_issue_key(repo)
    return dict(
        associated_issue=fetch_issue(issue_key, jira_url, jira_username, jira_token)
    ) if issue_key else None



"""
Fix issues from code review report
"""
import json
import logging
from pathlib import Path
from typing import Optional

import git
import typer
from microcore import ui

from ..cli_base import app
from ..constants import JSON_REPORT_FILE_NAME
from ..report_struct import Report, Issue


@app.command(help="Fix an issue from the code review report")
def fix(
    issue_number: int = typer.Argument(..., help="Issue number to fix"),
    report_path: Optional[str] = typer.Option(
        None,
        "--report",
        "-r",
        help="Path to the code review report (default: code-review-report.json)"
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", "-d", help="Only print changes without applying them"
    ),
    commit: bool = typer.Option(default=False, help="Commit changes after applying them"),
    push: bool = typer.Option(default=False, help="Push changes to the remote repository"),
) -> list[str]:
    """
    Applies the proposed change for the specified issue number from the code review report.
    """
    # Load the report
    report_path = report_path or JSON_REPORT_FILE_NAME
    try:
        report = Report.load(report_path)
    except (FileNotFoundError, json.JSONDecodeError) as e:
        logging.error(f"Failed to load report from {report_path}: {e}")
        raise typer.Exit(code=1)

    # Find the issue by number
    issue: Optional[Issue] = None
    for file_issues in report.issues.values():
        for i in file_issues:
            if i.id == issue_number:
                issue = i
                break
        if issue:
            break

    if not issue:
        logging.error(f"Issue #{issue_number} not found in the report")
        raise typer.Exit(code=1)

    if not issue.affected_lines:
        logging.error(f"Issue #{issue_number} has no affected lines specified")
        raise typer.Exit(code=1)

    if not any(affected_line.proposal for affected_line in issue.affected_lines):
        logging.error(f"Issue #{issue_number} has no proposal for fixing")
        raise typer.Exit(code=1)

    # Apply the fix
    logging.info(f"Fixing issue #{issue_number}: {ui.cyan(issue.title)}")

    for affected_line in issue.affected_lines:
        if not affected_line.proposal:
            continue

        file_path = Path(issue.file)
        if not file_path.exists():
            logging.error(f"File {file_path} not found")
            continue

        try:
            with open(file_path, "r", encoding="utf-8") as f:
                lines = f.readlines()
        except Exception as e:
            logging.error(f"Failed to read file {file_path}: {e}")
            continue

        # Check if line numbers are valid
        if affected_line.start_line < 1 or affected_line.end_line > len(lines):
            logging.error(
                f"Invalid line range: {affected_line.start_line}-{affected_line.end_line} "
                f"(file has {len(lines)} lines)"
            )
            continue

        # Get the affected line content for display
        affected_content = "".join(lines[affected_line.start_line - 1:affected_line.end_line])
        print(f"\nFile: {ui.blue(issue.file)}")
        print(f"Lines: {affected_line.start_line}-{affected_line.end_line}")
        print(f"Current content:\n{ui.red(affected_content)}")
        print(f"Proposed change:\n{ui.green(affected_line.proposal)}")

        if dry_run:
            print(f"{ui.yellow('Dry run')}: Changes not applied")
            continue

        # Apply the change
        proposal_lines = affected_line.proposal.splitlines(keepends=True)
        if not proposal_lines:
            proposal_lines = [""]
        elif not proposal_lines[-1].endswith(("\n", "\r")):
            # Ensure the last line has a newline if the original does
            if (
                affected_line.end_line < len(lines)
                and lines[affected_line.end_line - 1].endswith(("\n", "\r"))
            ):
                proposal_lines[-1] += "\n"

        lines[affected_line.start_line - 1:affected_line.end_line] = proposal_lines

        # Write changes back to the file
        try:
            with open(file_path, "w", encoding="utf-8") as f:
                f.writelines(lines)
            print(f"{ui.green('Success')}: Changes applied to {file_path}")
        except Exception as e:
            logging.error(f"Failed to write changes to {file_path}: {e}")
            raise typer.Exit(code=1)

    print(f"\n{ui.green('✓')} Issue #{issue_number} fixed successfully")

    changed_files = [file_path.as_posix()]
    if commit:
        commit_changes(
            changed_files,
            commit_message=f"[AI] Fix issue {issue_number}:{issue.title}",
            push=push
        )
    return changed_files


def commit_changes(
    files: list[str],
    repo: git.Repo = None,
    commit_message: str = "fix by AI",
    push: bool = True
) -> None:
    if opened_repo := not repo:
        repo = git.Repo(".")
    for i in files:
        repo.index.add(i)
    repo.index.commit(commit_message)
    if push:
        origin = repo.remotes.origin
        origin.push()
        logging.info(f"Changes pushed to {origin.name}")
    else:
        logging.info("Changes committed but not pushed to remote")
    if opened_repo:
        repo.close()



from pathlib import Path
from .env import Env

PROJECT_GITO_FOLDER = ".gito"
PROJECT_CONFIG_FILE_NAME = "config.toml"
PROJECT_CONFIG_FILE_PATH = Path(".gito") / PROJECT_CONFIG_FILE_NAME
PROJECT_CONFIG_BUNDLED_DEFAULTS_FILE = Path(__file__).resolve().parent / PROJECT_CONFIG_FILE_NAME
HOME_ENV_PATH = Path("~/.gito/.env").expanduser()
JSON_REPORT_FILE_NAME = "code-review-report.json"
GITHUB_MD_REPORT_FILE_NAME = "code-review-report.md"
EXECUTABLE = "gito"
TEXT_ICON_URL = 'https://raw.githubusercontent.com/Nayjest/Gito/main/press-kit/logo/gito-bot-1_64top.png'  # noqa: E501
HTML_TEXT_ICON = f'<a href="https://github.com/Nayjest/Gito"><img src="{TEXT_ICON_URL}" align="left" width=64 height=50 title="Gito v{Env.gito_version}"/></a>'  # noqa: E501
HTML_CR_COMMENT_MARKER = '<!-- GITO_COMMENT:CODE_REVIEW_REPORT -->'



from dataclasses import dataclass, field
from typing import Iterable, TYPE_CHECKING

from unidiff.patch import PatchSet, PatchedFile
from git import Repo


if TYPE_CHECKING:
    from .project_config import ProjectConfig
    from .report_struct import Report


@dataclass
class Context:
    report: "Report"
    config: "ProjectConfig"
    diff: PatchSet | Iterable[PatchedFile]
    repo: Repo
    pipeline_out: dict = field(default_factory=dict)



import os
import logging

import requests
import git
from fastcore.basics import AttrDict  # objects returned by ghapi
from ghapi.core import GhApi

from .project_config import ProjectConfig
from .utils import extract_gh_owner_repo


def gh_api(
    repo: git.Repo = None,  # used to resolve owner/repo
    config: ProjectConfig | None = None,  # used to resolve owner/repo
    token: str | None = None
) -> GhApi:
    if repo:
        # resolve owner/repo from repo.remotes.origin.url
        owner, repo_name = extract_gh_owner_repo(repo)
    else:
        if not config:
            config = ProjectConfig.load()
        # resolve owner/repo from github env vars (github actions)
        gh_env = config.prompt_vars.get("github_env", {})
        gh_repo = gh_env.get("github_repo")
        if not gh_repo:
            raise ValueError("GitHub repository not specified and not found in project config.")
        parts = gh_repo.split('/')
        if len(parts) != 2:
            raise ValueError(f"Invalid GitHub repository format: {gh_repo}. Expected 'owner/repo'.")
        owner, repo_name = parts

    token = resolve_gh_token(token)
    api = GhApi(owner, repo_name, token=token)
    return api


def resolve_gh_token(token_or_none: str | None = None) -> str | None:
    return token_or_none or os.getenv("GITHUB_TOKEN", None) or os.getenv("GH_TOKEN", None)


def post_gh_comment(
    gh_repository: str,  # e.g. "owner/repo"
    pr_or_issue_number: int,
    gh_token: str,
    text: str,
) -> bool:
    """
    Post a comment to a GitHub pull request or issue.
    Arguments:
        gh_repository (str): The GitHub repository in the format "owner/repo".
        pr_or_issue_number (int): The pull request or issue number.
        gh_token (str): GitHub personal access token with permissions to post comments.
        text (str): The comment text to post.
    Returns:
        True if the comment was posted successfully, False otherwise.
    """
    api_url = f"https://api.github.com/repos/{gh_repository}/issues/{pr_or_issue_number}/comments"
    headers = {
        "Authorization": f"token {gh_token}",
        "Accept": "application/vnd.github+json",
    }
    data = {"body": text}

    resp = requests.post(api_url, headers=headers, json=data)
    if 200 <= resp.status_code < 300:
        logging.info(f"Posted review comment to #{pr_or_issue_number} in {gh_repository}")
        return True
    else:
        logging.error(f"Failed to post comment: {resp.status_code} {resp.reason}\n{resp.text}")
        return False


def hide_gh_comment(
    comment: dict | str,
    token: str = None,
    reason: str = "OUTDATED"
) -> bool:
    """
    Hide a GitHub comment using GraphQL API with specified reason.
    Args:
        comment (dict | str):
            The comment to hide,
            either as a object returned from ghapi or a string node ID.
            note: comment.id is not the same as node_id.
        token (str): GitHub personal access token with permissions to minimize comments.
        reason (str): The reason for hiding the comment, e.g., "OUTDATED".
    """
    comment_node_id = comment.node_id if isinstance(comment, AttrDict) else comment
    token = resolve_gh_token(token)
    mutation = """
    mutation($commentId: ID!, $reason: ReportedContentClassifiers!) {
        minimizeComment(input: {subjectId: $commentId, classifier: $reason}) {
            minimizedComment { isMinimized }
        }
    }"""

    response = requests.post(
        "https://api.github.com/graphql",
        headers={"Authorization": f"Bearer {token}"},
        json={
            "query": mutation,
            "variables": {"commentId": comment_node_id, "reason": reason}
        }
    )
    success = (
        response.status_code == 200
        and response.json().get("data", {}).get("minimizeComment") is not None
    )
    if not success:
        logging.error(
            f"Failed to hide comment {comment_node_id}: "
            f"{response.status_code} {response.reason}\n{response.text}"
        )
    return success



import os
import fnmatch
import logging
from typing import Iterable
from pathlib import Path
from functools import partial

import microcore as mc
from gito.gh_api import gh_api
from microcore import ui
from git import Repo, Commit
from git.exc import GitCommandError
from unidiff import PatchSet, PatchedFile
from unidiff.constants import DEV_NULL

from .context import Context
from .project_config import ProjectConfig
from .report_struct import Report
from .constants import JSON_REPORT_FILE_NAME
from .utils import make_streaming_function
from .pipeline import Pipeline
from .env import Env


def review_subject_is_index(what):
    return not what or what == 'INDEX'


def is_binary_file(repo: Repo, file_path: str) -> bool:
    """
    Check if a file is binary by attempting to read it as text.
    Returns True if the file is binary, False otherwise.
    """
    try:
        # Attempt to read the file content from the repository tree
        content = repo.tree()[file_path].data_stream.read()
        # Try decoding as UTF-8; if it fails, it's likely binary
        content.decode("utf-8")
        return False
    except KeyError:
        try:
            fs_path = Path(repo.working_tree_dir) / file_path
            fs_path.read_text(encoding='utf-8')
            return False
        except FileNotFoundError:
            logging.error(f"File {file_path} not found in the repository.")
            return True
        except UnicodeDecodeError:
            return True
        except Exception as e:
            logging.error(f"Error reading file {file_path}: {e}")
            return True
    except UnicodeDecodeError:
        return True
    except Exception as e:
        logging.warning(f"Error checking if file {file_path} is binary: {e}")
        return True  # Conservatively treat errors as binary to avoid issues


def commit_in_branch(repo: Repo, commit: Commit, target_branch: str) -> bool:
    try:
        # exit code 0 if commit is ancestor of branch
        repo.git.merge_base('--is-ancestor', commit.hexsha, target_branch)
        return True
    except GitCommandError:
        pass
    return False


def get_base_branch(repo: Repo, pr: int | str = None):
    if os.getenv('GITHUB_ACTIONS'):

        # triggered from PR
        if base_ref := os.getenv('GITHUB_BASE_REF'):
            logging.info(f"Using GITHUB_BASE_REF:{base_ref} as base branch")
            return f'origin/{base_ref}'
        logging.info("GITHUB_BASE_REF is not available...")
        if pr:
            api = gh_api(repo=repo)
            pr_obj = api.pulls.get(pr)
            logging.info(
                f"Using 'origin/{pr_obj.base.ref}' as base branch "
                f"(received via GH API for PR#{pr})"
            )
            return f'origin/{pr_obj.base.ref}'

    try:
        logging.info(
            "Trying to resolve base branch from repo.remotes.origin.refs.HEAD.reference.name..."
        )
        # 'origin/main', 'origin/master', etc
        # Stopped working in github actions since 07/2025
        return repo.remotes.origin.refs.HEAD.reference.name
    except AttributeError:
        try:
            logging.info(
                "Checking if repo has 'main' or 'master' branchs to use as --against branch..."
            )
            remote_refs = repo.remotes.origin.refs
            for branch_name in ['main', 'master']:
                if hasattr(remote_refs, branch_name):
                    return f'origin/{branch_name}'
        except Exception:
            pass

        logging.error("Could not determine default branch from remote refs.")
        raise ValueError("No default branch found in the repository.")


def get_diff(
    repo: Repo = None,
    what: str = None,
    against: str = None,
    use_merge_base: bool = True,
    pr: str | int = None
) -> PatchSet | list[PatchedFile]:
    repo = repo or Repo(".")
    if not against:
        # 'origin/main', 'origin/master', etc
        against = get_base_branch(repo, pr=pr)
    if review_subject_is_index(what):
        what = None  # working copy
    if use_merge_base:
        if review_subject_is_index(what):
            try:
                current_ref = repo.active_branch.name
            except TypeError:
                # In detached HEAD state, use HEAD directly
                current_ref = "HEAD"
                logging.info(
                    "Detected detached HEAD state, using HEAD as current reference"
                )
        else:
            current_ref = what
        merge_base = repo.merge_base(current_ref or repo.active_branch.name, against)[0]
        logging.info(
            f"Merge base({ui.green(current_ref)},{ui.yellow(against)})"
            f" --> {ui.cyan(merge_base.hexsha)}"
        )
        # if branch is already an ancestor of "against", merge_base == branch ⇒ it’s been merged
        if merge_base.hexsha == repo.commit(current_ref or repo.active_branch.name).hexsha:
            # @todo: check case: reviewing working copy index in main branch #103
            logging.info(
                f"Branch is already merged. ({ui.green(current_ref)} vs {ui.yellow(against)})"
            )
            merge_sha = repo.git.log(
                '--merges',
                '--ancestry-path',
                f'{current_ref}..{against}',
                '-n',
                '1',
                '--pretty=format:%H'
            ).strip()
            if merge_sha:
                logging.info(f"Merge commit is {ui.cyan(merge_sha)}")
                merge_commit = repo.commit(merge_sha)

                other_merge_parent = None
                for parent in merge_commit.parents:
                    logging.info(f"Checking merge parent: {parent.hexsha[:8]}")
                    if parent.hexsha == merge_base.hexsha:
                        logging.info(f"merge parent is {ui.cyan(parent.hexsha[:8])}, skipping")
                        continue
                    if not commit_in_branch(repo, parent, against):
                        logging.warning(f"merge parent is not in {against}, skipping")
                        continue
                    logging.info(f"Found other merge parent: {ui.cyan(parent.hexsha[:8])}")
                    other_merge_parent = parent
                    break
                if other_merge_parent:
                    first_common_ancestor = repo.merge_base(other_merge_parent, merge_base)[0]
                    # for gito remote (feature_branch vs origin/main)
                    # the same merge base appears in first_common_ancestor again
                    if first_common_ancestor.hexsha == merge_base.hexsha:
                        if merge_base.parents:
                            first_common_ancestor = repo.merge_base(
                                other_merge_parent, merge_base.parents[0]
                            )[0]
                        else:
                            logging.error(
                                "merge_base has no parents, "
                                "using merge_base as first_common_ancestor"
                            )
                    logging.info(
                        f"{what} will be compared to "
                        f"first common ancestor of {what} and {against}: "
                        f"{ui.cyan(first_common_ancestor.hexsha[:8])}"
                    )
                    against = first_common_ancestor.hexsha
                else:
                    logging.error(f"Can't find other merge parent for {merge_sha}")
            else:
                logging.error(
                    f"No merge‐commit found for {current_ref!r}→{against!r}; "
                    "falling back to merge‐base diff"
                )
        else:
            # normal case: branch not yet merged
            against = merge_base.hexsha
            logging.info(
                f"Using merge base: {ui.cyan(merge_base.hexsha[:8])} ({merge_base.summary})"
            )
    logging.info(
        f"Making diff: {ui.green(what or 'INDEX')} vs {ui.yellow(against)}"
    )
    diff_content = repo.git.diff(against, what)
    diff = PatchSet.from_string(diff_content)

    # Filter out binary files
    non_binary_diff = PatchSet([])
    for patched_file in diff:
        # Check if the file is binary using the source or target file path
        file_path = (
            patched_file.target_file
            if patched_file.target_file != DEV_NULL
            else patched_file.source_file
        )
        if file_path == DEV_NULL:
            continue
        if is_binary_file(repo, file_path.removeprefix("b/")):
            logging.info(f"Skipping binary file: {patched_file.path}")
            continue
        non_binary_diff.append(patched_file)
    return non_binary_diff


def filter_diff(
    patch_set: PatchSet | Iterable[PatchedFile], filters: str | list[str]
) -> PatchSet | Iterable[PatchedFile]:
    """
    Filter the diff files by the given fnmatch filters.
    """
    assert isinstance(filters, (list, str))
    if not isinstance(filters, list):
        filters = [f.strip() for f in filters.split(",") if f.strip()]
    if not filters:
        return patch_set
    files = [
        file
        for file in patch_set
        if any(fnmatch.fnmatch(file.path, pattern) for pattern in filters)
    ]
    return files


def file_lines(repo: Repo, file: str, max_tokens: int = None, use_local_files: bool = False) -> str:
    if use_local_files:
        file_path = Path(repo.working_tree_dir) / file
        try:
            text = file_path.read_text(encoding='utf-8')
        except (FileNotFoundError, UnicodeDecodeError) as e:
            logging.warning(f"Could not read file {file} from working directory: {e}")
            text = repo.tree()[file].data_stream.read().decode('utf-8')
    else:
        # Read from HEAD (committed version)
        text = repo.tree()[file].data_stream.read().decode('utf-8')

    lines = [f"{i + 1}: {line}\n" for i, line in enumerate(text.splitlines())]
    if max_tokens:
        lines, removed_qty = mc.tokenizing.fit_to_token_size(lines, max_tokens)
        if removed_qty:
            lines.append(
                f"(!) DISPLAYING ONLY FIRST {len(lines)} LINES DUE TO LARGE FILE SIZE\n"
            )
    return "".join(lines)


def make_cr_summary(ctx: Context, **kwargs) -> str:
    return (
        mc.prompt(
            ctx.config.summary_prompt,
            diff=mc.tokenizing.fit_to_token_size(ctx.diff, ctx.config.max_code_tokens)[0],
            issues=ctx.report.issues,
            pipeline_out=ctx.pipeline_out,
            env=Env,
            **ctx.config.prompt_vars,
            **kwargs,
        ).to_llm()
        if ctx.config.summary_prompt
        else ""
    )


class NoChangesInContextError(Exception):
    """
    Exception raised when there are no changes in the context to review /answer questions.
    """


def _prepare(
    repo: Repo = None,
    what: str = None,
    against: str = None,
    filters: str | list[str] = "",
    use_merge_base: bool = True,
    pr: str | int = None,
):
    repo = repo or Repo(".")
    cfg = ProjectConfig.load_for_repo(repo)
    diff = get_diff(
        repo=repo, what=what, against=against, use_merge_base=use_merge_base, pr=pr,
    )
    diff = filter_diff(diff, filters)
    if not diff:
        raise NoChangesInContextError()
    lines = {
        file_diff.path: (
            file_lines(
                repo,
                file_diff.path,
                cfg.max_code_tokens
                - mc.tokenizing.num_tokens_from_string(str(file_diff)),
                use_local_files=review_subject_is_index(what)
            )
            if file_diff.target_file != DEV_NULL and not file_diff.is_added_file
            else ""
        )
        for file_diff in diff
    }
    return repo, cfg, diff, lines


def get_affected_code_block(repo: Repo, file: str, start_line: int, end_line: int) -> str | None:
    if not start_line or not end_line:
        return None
    try:
        if isinstance(start_line, str):
            start_line = int(start_line)
        if isinstance(end_line, str):
            end_line = int(end_line)
        lines = file_lines(repo, file, max_tokens=None, use_local_files=True)
        if lines:
            lines = [""] + lines.splitlines()
            return "\n".join(
                lines[start_line: end_line + 1]
            )
    except Exception as e:
        logging.error(
            f"Error getting affected code block for {file} from {start_line} to {end_line}: {e}"
        )
    return None


def provide_affected_code_blocks(issues: dict, repo: Repo):
    for file, file_issues in issues.items():
        for issue in file_issues:
            for i in issue.get("affected_lines", []):
                file_name = i.get("file", issue.get("file", file))
                if block := get_affected_code_block(
                    repo,
                    file_name,
                    i.get("start_line"),
                    i.get("end_line")
                ):
                    i["affected_code"] = block


async def review(
    repo: Repo = None,
    what: str = None,
    against: str = None,
    filters: str | list[str] = "",
    use_merge_base: bool = True,
    out_folder: str | os.PathLike | None = None,
    pr: str | int = None
):
    try:
        repo, cfg, diff, lines = _prepare(
            repo=repo,
            what=what,
            against=against,
            filters=filters,
            use_merge_base=use_merge_base,
            pr=pr,
        )
    except NoChangesInContextError:
        logging.error("No changes to review")
        return
    responses = await mc.llm_parallel(
        [
            mc.prompt(
                cfg.prompt,
                input=file_diff,
                file_lines=lines[file_diff.path],
                **cfg.prompt_vars,
            )
            for file_diff in diff
        ],
        retries=cfg.retries,
        parse_json=True,
    )
    issues = {file.path: issues for file, issues in zip(diff, responses) if issues}
    provide_affected_code_blocks(issues, repo)
    exec(cfg.post_process, {"mc": mc, **locals()})
    out_folder = Path(out_folder or repo.working_tree_dir)
    out_folder.mkdir(parents=True, exist_ok=True)
    report = Report(issues=issues, number_of_processed_files=len(diff))
    ctx = Context(
        report=report,
        config=cfg,
        diff=diff,
        repo=repo,
    )
    if cfg.pipeline_steps:
        pipe = Pipeline(
            ctx=ctx,
            steps=cfg.pipeline_steps
        )
        pipe.run()
    else:
        logging.info("No pipeline steps defined, skipping pipeline execution")

    report.summary = make_cr_summary(ctx)
    report.save(file_name=out_folder / JSON_REPORT_FILE_NAME)
    report_text = report.render(cfg, Report.Format.MARKDOWN)
    text_report_path = out_folder / "code-review-report.md"
    text_report_path.write_text(report_text, encoding="utf-8")
    report.to_cli()


def answer(
    question: str,
    repo: Repo = None,
    what: str = None,
    against: str = None,
    filters: str | list[str] = "",
    use_merge_base: bool = True,
    use_pipeline: bool = True,
    prompt_file: str = None,
    pr: str | int = None,
) -> str | None:
    try:
        repo, config, diff, lines = _prepare(
            repo=repo,
            what=what,
            against=against,
            filters=filters,
            use_merge_base=use_merge_base,
            pr=pr
        )
    except NoChangesInContextError:
        logging.error("No changes to review")
        return

    ctx = Context(
        repo=repo,
        diff=diff,
        config=config,
        report=Report()
    )
    if use_pipeline:
        pipe = Pipeline(
            ctx=ctx,
            steps=config.pipeline_steps
        )
        pipe.run()
    if prompt_file:
        prompt_func = partial(mc.tpl, prompt_file)
    else:
        prompt_func = partial(mc.prompt, config.answer_prompt)
    response = mc.llm(
        prompt_func(
            question=question,
            diff=diff,
            all_file_lines=lines,
            pipeline_out=ctx.pipeline_out,
            **config.prompt_vars,
        ),
        callback=make_streaming_function() if Env.verbosity == 0 else None,
    )
    return response



"""
Fix issues from code review report
"""

import logging
import os
import re
from pathlib import Path
from typing import Optional
import zipfile

import requests
import typer
from fastcore.basics import AttrDict
from microcore import ui
from ghapi.all import GhApi
import git

from ..cli_base import app
from ..constants import JSON_REPORT_FILE_NAME, HTML_TEXT_ICON
from ..core import answer
from ..gh_api import post_gh_comment, resolve_gh_token
from ..project_config import ProjectConfig
from ..utils import extract_gh_owner_repo
from .fix import fix


def cleanup_comment_addressed_to_gito(text):
    if not text:
        return text
    patterns = [
        r'^\s*gito\b',
        r'^\s*ai\b',
        r'^\s*bot\b',
        r'^\s*@gito\b',
        r'^\s*@ai\b',
        r'^\s*@bot\b'
    ]
    result = text
    # Remove each pattern from the beginning
    for pattern in patterns:
        result = re.sub(pattern, '', result, flags=re.IGNORECASE)

    # Remove leading comma and spaces that may be left after prefix removal
    result = re.sub(r'^\s*,\s*', '', result)

    # Clean up extra whitespace
    result = re.sub(r'\s+', ' ', result).strip()
    return result


@app.command(hidden=True)
def react_to_comment(
    comment_id: int = typer.Argument(),
    gh_token: str = typer.Option(
        "",
        "--gh-token",
        "--token",
        "-t",
        "--github-token",
        help="GitHub token for authentication",
    ),
    dry_run: bool = typer.Option(
        False, "--dry-run", "-d", help="Only print changes without applying them"
    ),
):
    """
    Handles direct agent instructions from pull request comments.

    Note: Not for local usage. Designed for execution within GitHub Actions workflows.

    Fetches the PR comment by ID, parses agent directives, and executes the requested
    actions automatically to enable seamless code review workflow integration.
    """
    repo = git.Repo(".")  # Current directory
    owner, repo_name = extract_gh_owner_repo(repo)
    logging.info(f"Using repository: {ui.yellow}{owner}/{repo_name}{ui.reset}")
    gh_token = resolve_gh_token(gh_token)
    api = GhApi(owner=owner, repo=repo_name, token=gh_token)
    comment = api.issues.get_comment(comment_id=comment_id)
    logging.info(
        f"Comment by {ui.yellow('@' + comment.user.login)}: "
        f"{ui.green(comment.body)}\n"
        f"url: {comment.html_url}"
    )

    cfg = ProjectConfig.load_for_repo(repo)
    if not any(
        trigger.lower() in comment.body.lower() for trigger in cfg.mention_triggers
    ):
        ui.error("No mention trigger found in comment, no reaction added.")
        return
    try:
        logging.info("Comment contains mention trigger, reacting with 'eyes'.")
        api.reactions.create_for_issue_comment(comment_id=comment_id, content="eyes")
    except Exception as e:
        logging.error("Error reacting to comment with emoji: %s", str(e))
    pr = int(comment.issue_url.split("/")[-1])
    print(f"Processing comment for PR #{pr}...")

    issue_ids = extract_fix_args(comment.body)
    if issue_ids:
        logging.info(f"Extracted issue IDs: {ui.yellow(str(issue_ids))}")
        out_folder = "artifact"
        download_latest_code_review_artifact(
            api, pr_number=pr, gh_token=gh_token, out_folder=out_folder
        )
        fix(
            issue_ids[0],  # @todo: support multiple IDs
            report_path=Path(out_folder) / JSON_REPORT_FILE_NAME,
            dry_run=dry_run,
            commit=not dry_run,
            push=not dry_run,
        )
        logging.info("Fix applied successfully.")
    elif is_review_request(comment.body):
        ref = repo.active_branch.name
        logging.info(f"Triggering code-review workflow, ref='{ref}'")
        api.actions.create_workflow_dispatch(
            workflow_id="gito-code-review.yml",
            ref=ref,
            inputs={"pr_number": str(pr)},
        )
    else:
        if cfg.answer_github_comments:
            question = cleanup_comment_addressed_to_gito(comment.body)
            response = answer(question, repo=repo, pr=pr)
            post_gh_comment(
                gh_repository=f"{owner}/{repo_name}",
                pr_or_issue_number=pr,
                gh_token=gh_token,
                text=HTML_TEXT_ICON+response,
            )
        else:
            ui.error("Can't identify target command in the text.")
            return


def last_code_review_run(api: GhApi, pr_number: int) -> AttrDict | None:
    pr = api.pulls.get(pr_number)
    sha = pr["head"]["sha"]  # noqa
    branch = pr["head"]["ref"]

    runs = api.actions.list_workflow_runs_for_repo(branch=branch)["workflow_runs"]
    # Find the run for this SHA
    run = next(
        (
            r
            for r in runs  # r['head_sha'] == sha and
            if (
                any(
                    marker in r["path"].lower()
                    for marker in ["code-review", "code_review", "cr"]
                )
                or "gito.yml" in r["name"].lower()
            )
            and r["status"] == "completed"
        ),
        None,
    )
    return run


def download_latest_code_review_artifact(
    api: GhApi, pr_number: int, gh_token: str, out_folder: Optional[str] = "artifact"
) -> tuple[str, dict] | None:
    run = last_code_review_run(api, pr_number)
    if not run:
        raise Exception("No workflow run found for this PR/SHA")

    artifacts = api.actions.list_workflow_run_artifacts(run["id"])["artifacts"]
    if not artifacts:
        raise Exception("No artifacts found for this workflow run")

    latest_artifact = artifacts[0]
    url = latest_artifact["archive_download_url"]
    print(f"Artifact: {latest_artifact['name']}, Download URL: {url}")
    headers = {"Authorization": f"token {gh_token}"} if gh_token else {}
    zip_path = "artifact.zip"
    try:
        with requests.get(url, headers=headers, stream=True) as r:
            r.raise_for_status()
            with open(zip_path, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)

        # Unpack to ./artifact
        os.makedirs("artifact", exist_ok=True)
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall("artifact")
    finally:
        if os.path.exists(zip_path):
            os.remove(zip_path)

    print("Artifact unpacked to ./artifact")


def extract_fix_args(text: str) -> list[int]:
    pattern1 = r"fix\s+(?:issues?)?(?:\s+)?#?(\d+(?:\s*,\s*#?\d+)*)"
    match = re.search(pattern1, text)
    if match:
        numbers_str = match.group(1)
        numbers = re.findall(r"\d+", numbers_str)
        issue_numbers = [int(num) for num in numbers]
        return issue_numbers
    return []


def is_review_request(text: str) -> bool:
    text = text.lower().strip()
    trigger_words = ['review', 'run', 'code-review']
    if any(f"/{word}" in text for word in trigger_words):
        return True
    parts = text.split()
    if len(parts) == 2 and parts[1] in trigger_words:
        return True
    return False



import logging
import os
from time import sleep

import typer
from ghapi.core import GhApi

from ..cli_base import app
from ..constants import GITHUB_MD_REPORT_FILE_NAME, HTML_CR_COMMENT_MARKER
from ..gh_api import (
    post_gh_comment,
    resolve_gh_token,
    hide_gh_comment,
)
from ..project_config import ProjectConfig


@app.command(name="github-comment", help="Leave a GitHub PR comment with the review.")
def post_github_cr_comment(
    md_report_file: str = typer.Option(default=None),
    pr: int = typer.Option(default=None),
    gh_repo: str = typer.Option(default=None, help="owner/repo"),
    token: str = typer.Option(
        "", help="GitHub token (or set GITHUB_TOKEN env var)"
    ),
):
    """
    Leaves a comment with the review on the current GitHub pull request.
    """
    file = md_report_file or GITHUB_MD_REPORT_FILE_NAME
    if not os.path.exists(file):
        logging.error(f"Review file not found: {file}, comment will not be posted.")
        raise typer.Exit(4)

    with open(file, "r", encoding="utf-8") as f:
        body = f.read()

    token = resolve_gh_token(token)
    if not token:
        print("GitHub token is required (--token or GITHUB_TOKEN env var).")
        raise typer.Exit(1)
    config = ProjectConfig.load()
    gh_env = config.prompt_vars["github_env"]
    gh_repo = gh_repo or gh_env.get("github_repo", "")
    pr_env_val = gh_env.get("github_pr_number", "")
    logging.info(f"github_pr_number = {pr_env_val}")

    if not pr:
        # e.g. could be "refs/pull/123/merge" or a direct number
        if "/" in pr_env_val and "pull" in pr_env_val:
            # refs/pull/123/merge
            try:
                pr_num_candidate = pr_env_val.strip("/").split("/")
                idx = pr_num_candidate.index("pull")
                pr = int(pr_num_candidate[idx + 1])
            except Exception:
                pass
        else:
            try:
                pr = int(pr_env_val)
            except ValueError:
                pass
    if not pr:
        if pr_str := os.getenv("PR_NUMBER_FROM_WORKFLOW_DISPATCH"):
            try:
                pr = int(pr_str)
            except ValueError:
                pass
    if not pr:
        logging.error("Could not resolve PR number from environment variables.")
        raise typer.Exit(3)

    if not post_gh_comment(gh_repo, pr, token, body):
        raise typer.Exit(5)

    if config.collapse_previous_code_review_comments:
        sleep(1)
        collapse_gh_outdated_cr_comments(gh_repo, pr, token)


def collapse_gh_outdated_cr_comments(
    gh_repository: str,
    pr_or_issue_number: int,
    token: str = None
):
    """
    Collapse outdated code review comments in a GitHub pull request or issue.
    """
    logging.info(f"Collapsing outdated comments in {gh_repository} #{pr_or_issue_number}...")

    token = resolve_gh_token(token)
    owner, repo = gh_repository.split('/')
    api = GhApi(owner, repo, token=token)

    comments = api.issues.list_comments(pr_or_issue_number)
    review_marker = HTML_CR_COMMENT_MARKER
    collapsed_title = "🗑️ Outdated Code Review by Gito"
    collapsed_marker = f"<summary>{collapsed_title}</summary>"
    outdated_comments = [
        c for c in comments
        if c.body and review_marker in c.body and collapsed_marker not in c.body
    ][:-1]
    if not outdated_comments:
        logging.info("No outdated comments found")
        return
    for comment in outdated_comments:
        logging.info(f"Collapsing comment {comment.id}...")
        new_body = f"<details>\n<summary>{collapsed_title}</summary>\n\n{comment.body}\n</details>"
        api.issues.update_comment(comment.id, new_body)
        hide_gh_comment(comment.node_id, token)
    logging.info("All outdated comments collapsed successfully.")



import re
import sys
import os
from pathlib import Path
import importlib.metadata
from typing import Optional

import typer
import git
from git import Repo


_EXT_TO_HINT: dict[str, str] = {
    # scripting & languages
    ".py": "python",
    ".js": "javascript",
    ".ts": "typescript",
    ".java": "java",
    ".c": "c",
    ".cpp": "cpp",
    ".cc": "cpp",
    ".cxx": "cpp",
    ".h": "cpp",
    ".hpp": "cpp",
    ".cs": "csharp",
    ".rb": "ruby",
    ".go": "go",
    ".rs": "rust",
    ".swift": "swift",
    ".kt": "kotlin",
    ".scala": "scala",
    ".dart": "dart",
    ".php": "php",
    ".pl": "perl",
    ".pm": "perl",
    ".lua": "lua",
    # web & markup
    ".html": "html",
    ".htm": "html",
    ".css": "css",
    ".scss": "scss",
    ".less": "less",
    ".json": "json",
    ".xml": "xml",
    ".yaml": "yaml",
    ".yml": "yaml",
    ".toml": "toml",
    ".ini": "ini",
    ".csv": "csv",
    ".md": "markdown",
    ".rst": "rest",
    # shell & config
    ".sh": "bash",
    ".bash": "bash",
    ".zsh": "bash",
    ".fish": "bash",
    ".ps1": "powershell",
    ".dockerfile": "dockerfile",
    # build & CI
    ".makefile": "makefile",
    ".mk": "makefile",
    "CMakeLists.txt": "cmake",
    "Dockerfile": "dockerfile",
    ".gradle": "groovy",
    ".travis.yml": "yaml",
    # data & queries
    ".sql": "sql",
    ".graphql": "graphql",
    ".proto": "protobuf",
    ".yara": "yara",
}


def syntax_hint(file_path: str | Path) -> str:
    """
    Returns a syntax highlighting hint based on the file's extension or name.

    This can be used to annotate code blocks for rendering with syntax highlighting,
    e.g., using Markdown-style code blocks: ```<syntax_hint>\n<code>\n```.

    Args:
      file_path (str | Path): Path to the file.

    Returns:
      str: A syntax identifier suitable for code highlighting (e.g., 'python', 'json').
    """
    p = Path(file_path)
    ext = p.suffix.lower()
    if not ext:
        name = p.name.lower()
        if name == "dockerfile":
            return "dockerfile"
        return ""
    return _EXT_TO_HINT.get(ext, ext.lstrip("."))


def is_running_in_github_action():
    return os.getenv("GITHUB_ACTIONS") == "true"


def no_subcommand(app: typer.Typer) -> bool:
    """
    Checks if the current script is being invoked as a command in a target Typer application.
    """
    return not (
        (first_arg := next((a for a in sys.argv[1:] if not a.startswith('-')), None))
        and first_arg in (
            cmd.name or cmd.callback.__name__.replace('_', '-')
            for cmd in app.registered_commands
        )
        or '--help' in sys.argv
    )


def parse_refs_pair(refs: str) -> tuple[str | None, str | None]:
    SEPARATOR = '..'
    if not refs:
        return None, None
    if SEPARATOR not in refs:
        return refs, None
    what, against = refs.split(SEPARATOR, 1)
    return what or None, against or None


def max_line_len(text: str) -> int:
    return max((len(line) for line in text.splitlines()), default=0)


def block_wrap_lr(
    text: str,
    left: str = "",
    right: str = "",
    max_rwrap: int = 60,
    min_wrap: int = 0,
) -> str:
    ml = max(max_line_len(text), min_wrap)
    lines = text.splitlines()
    wrapped_lines = []
    for line in lines:
        ln = left+line
        if ml <= max_rwrap:
            ln += ' ' * (ml - len(line)) + right
        wrapped_lines.append(ln)
    return "\n".join(wrapped_lines)


def extract_gh_owner_repo(repo: git.Repo) -> tuple[str, str]:
    """
    Extracts the GitHub owner and repository name.

    Returns:
        tuple[str, str]: A tuple containing the owner and repository name.
    """
    remote_url = repo.remotes.origin.url
    if remote_url.startswith('git@github.com:'):
        # SSH format: git@github.com:owner/repo.git
        repo_path = remote_url.split(':')[1].replace('.git', '')
    elif remote_url.startswith('https://github.com/'):
        # HTTPS format: https://github.com/owner/repo.git
        repo_path = remote_url.replace('https://github.com/', '').replace('.git', '')
    else:
        raise ValueError("Unsupported remote URL format")
    owner, repo_name = repo_path.split('/')
    return owner, repo_name


def detect_github_env() -> dict:
    """
    Try to detect GitHub repository/PR info from environment variables (for GitHub Actions).
    Returns a dict with github_repo, github_pr_sha, github_pr_number, github_ref, etc.
    """
    repo = os.environ.get("GITHUB_REPOSITORY", "")
    pr_sha = os.environ.get("GITHUB_SHA", "")
    pr_number = os.environ.get("GITHUB_REF", "")
    branch = ""
    ref = os.environ.get("GITHUB_REF", "")
    # Try to resolve PR head SHA if available.
    # On PRs, GITHUB_HEAD_REF/BASE_REF contain branch names.
    if "GITHUB_HEAD_REF" in os.environ:
        branch = os.environ["GITHUB_HEAD_REF"]
    elif ref.startswith("refs/heads/"):
        branch = ref[len("refs/heads/"):]
    elif ref.startswith("refs/pull/"):
        # for pull_request events
        branch = ref

    d = {
        "github_repo": repo,
        "github_pr_sha": pr_sha,
        "github_pr_number": pr_number,
        "github_branch": branch,
        "github_ref": ref,
    }
    # Fallback for local usage: try to get from git
    if not repo:
        git_repo = None
        try:
            git_repo = Repo(".", search_parent_directories=True)
            origin = git_repo.remotes.origin.url
            # e.g. git@github.com:Nayjest/ai-code-review.git -> Nayjest/ai-code-review
            match = re.search(r"[:/]([\w\-]+)/([\w\-\.]+?)(\.git)?$", origin)
            if match:
                d["github_repo"] = f"{match.group(1)}/{match.group(2)}"
            d["github_pr_sha"] = git_repo.head.commit.hexsha
            d["github_branch"] = (
                git_repo.active_branch.name if hasattr(git_repo, "active_branch") else ""
            )
        except Exception:
            pass
        finally:
            if git_repo:
                try:
                    git_repo.close()
                except Exception:
                    pass
    # If branch is not a commit SHA, prefer branch for links
    if d["github_branch"]:
        d["github_pr_sha_or_branch"] = d["github_branch"]
    elif d["github_pr_sha"]:
        d["github_pr_sha_or_branch"] = d["github_pr_sha"]
    else:
        d["github_pr_sha_or_branch"] = "main"
    return d


def make_streaming_function(handler: Optional[callable] = None) -> callable:
    def stream(text):
        if handler:
            text = handler(text)
        print(text, end='', flush=True)
    return stream


def version() -> str:
    return importlib.metadata.version("gito.bot")


def remove_html_comments(text):
    """
    Removes all HTML comments (<!-- ... -->) from the input text.
    """
    return re.sub(r'<!--.*?-->\s*', '', text, flags=re.DOTALL)



import os
import sys
import logging

import typer
from git import Repo

from ..cli_base import app, arg_refs
from ..issue_trackers import resolve_issue_key

import requests


def post_linear_comment(issue_key, text, api_key):
    response = requests.post(
       'https://api.linear.app/graphql',
       headers={'Authorization': api_key, 'Content-Type': 'application/json'},
       json={
           'query': '''
               mutation($issueId: String!, $body: String!) {
                   commentCreate(input: {issueId: $issueId, body: $body}) {
                       comment { id }
                   }
               }
           ''',
           'variables': {'issueId': issue_key, 'body': text}
       }
    )
    return response.json()


@app.command()
def linear_comment(
    text: str = typer.Argument(None),
    refs: str = arg_refs(),
):
    if text is None or text == "-":
        # Read from stdin if no text provided
        text = sys.stdin.read()

    if not text or not text.strip():
        typer.echo("Error: No comment text provided.", err=True)
        raise typer.Exit(code=1)

    api_key = os.getenv("LINEAR_API_KEY")
    if not api_key:
        logging.error("LINEAR_API_KEY environment variable is not set")
        return

    repo = Repo(".")
    key = resolve_issue_key(repo)
    post_linear_comment(key, text, api_key)
    logging.info("Comment posted to Linear issue %s", key)



import os
import asyncio
import logging
import sys
import textwrap

import microcore as mc
import typer
from git import Repo

from .core import review, get_diff, filter_diff, answer
from .cli_base import (
    app,
    args_to_target,
    arg_refs,
    arg_what,
    arg_filters,
    arg_out,
    arg_against,
    get_repo_context,
)
from .report_struct import Report
from .constants import HOME_ENV_PATH, GITHUB_MD_REPORT_FILE_NAME
from .bootstrap import bootstrap
from .utils import no_subcommand, extract_gh_owner_repo, remove_html_comments
from .gh_api import resolve_gh_token

# Import fix command to register it
from .commands import fix, gh_react_to_comment, repl, deploy, version  # noqa
from .commands.gh_post_review_comment import post_github_cr_comment
from .commands.linear_comment import linear_comment

app_no_subcommand = typer.Typer(pretty_exceptions_show_locals=False)


def main():
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    # Help subcommand alias: if 'help' appears as first non-option arg, replace it with '--help'
    if len(sys.argv) > 1 and sys.argv[1] == "help":
        sys.argv = [sys.argv[0]] + sys.argv[2:] + ["--help"]

    if no_subcommand(app):
        bootstrap()
        app_no_subcommand()
    else:
        app()


@app.callback(invoke_without_command=True)
def cli(
    ctx: typer.Context,
    verbose: bool = typer.Option(default=None),
    verbosity: int = typer.Option(
        None,
        '--verbosity', '-v',
        help="Set verbosity level (0-3, default 1)"
    ),
):
    if verbose is not None and verbosity is not None:
        raise typer.BadParameter(
            "Please specify either --verbose or --verbosity, not both."
        )
    if verbose is not None:
        verbosity = 2 if verbose else 0
    if verbosity is None:
        verbosity = 1

    if ctx.invoked_subcommand != "setup":
        bootstrap(verbosity)


@app_no_subcommand.command(name="review", help="Perform code review")
@app.command(name="review", help="Perform code review")
@app.command(name="run", hidden=True)
def cmd_review(
    refs: str = arg_refs(),
    what: str = arg_what(),
    against: str = arg_against(),
    filters: str = arg_filters(),
    merge_base: bool = typer.Option(default=True, help="Use merge base for comparison"),
    url: str = typer.Option("", "--url", help="Git repository URL"),
    post_comment: bool = typer.Option(default=False, help="Post review comment to GitHub"),
    pr: int = typer.Option(
        default=None,
        help=textwrap.dedent("""\n
        GitHub Pull Request number to post the comment to
        (for local usage together with --post-comment,
        in the github actions PR is resolved from the environment)
        """)
    ),
    out: str = arg_out()
):
    _what, _against = args_to_target(refs, what, against)
    pr = pr or os.getenv("PR_NUMBER_FROM_WORKFLOW_DISPATCH")
    with get_repo_context(url, _what) as (repo, out_folder):
        asyncio.run(review(
            repo=repo,
            what=_what,
            against=_against,
            filters=filters,
            use_merge_base=merge_base,
            out_folder=out or out_folder,
            pr=pr,
        ))
        if post_comment:
            try:
                owner, repo_name = extract_gh_owner_repo(repo)
            except ValueError as e:
                logging.error(
                    "Error posting comment:\n"
                    "Could not extract GitHub owner and repository name from the local repository."
                )
                raise typer.Exit(code=1) from e
            post_github_cr_comment(
                md_report_file=os.path.join(out or out_folder, GITHUB_MD_REPORT_FILE_NAME),
                pr=pr,
                gh_repo=f"{owner}/{repo_name}",
                token=resolve_gh_token()
            )


@app.command(name="ask", help="Answer questions about codebase changes")
@app.command(name="answer", hidden=True)
@app.command(name="talk", hidden=True)
def cmd_answer(
    question: str = typer.Argument(help="Question to ask about the codebase changes"),
    refs: str = arg_refs(),
    what: str = arg_what(),
    against: str = arg_against(),
    filters: str = arg_filters(),
    merge_base: bool = typer.Option(default=True, help="Use merge base for comparison"),
    use_pipeline: bool = typer.Option(default=True),
    post_to: str = typer.Option(
        help="Post answer to ... Supported values: linear",
        default=None,
        show_default=False
    ),
    pr: int = typer.Option(
        default=None,
        help="GitHub Pull Request number"
    ),
):
    _what, _against = args_to_target(refs, what, against)
    pr = pr or os.getenv("PR_NUMBER_FROM_WORKFLOW_DISPATCH")
    if str(question).startswith("tpl:"):
        prompt_file = str(question)[4:]
        question = ""
    else:
        prompt_file = None
    out = answer(
        question=question,
        what=_what,
        against=_against,
        filters=filters,
        use_merge_base=merge_base,
        prompt_file=prompt_file,
        use_pipeline=use_pipeline,
        pr=pr,
    )
    if post_to == 'linear':
        logging.info("Posting answer to Linear...")
        linear_comment(remove_html_comments(out))
    return out


@app.command(help="Configure LLM for local usage interactively")
def setup():
    mc.interactive_setup(HOME_ENV_PATH)


@app.command(name="render")
@app.command(name="report", hidden=True)
def render(
    format: str = typer.Argument(default=Report.Format.CLI),
    source: str = typer.Option(
        "",
        "--src",
        "--source",
        help="Source file (json) to load the report from"
    )
):
    Report.load(file_name=source).to_cli(report_format=format)


@app.command(help="List files in the diff. Might be useful to check what will be reviewed.")
def files(
    refs: str = arg_refs(),
    what: str = arg_what(),
    against: str = arg_against(),
    filters: str = arg_filters(),
    merge_base: bool = typer.Option(default=True, help="Use merge base for comparison"),
    diff: bool = typer.Option(default=False, help="Show diff content")
):
    _what, _against = args_to_target(refs, what, against)
    repo = Repo(".")
    try:
        patch_set = get_diff(repo=repo, what=_what, against=_against, use_merge_base=merge_base)
        patch_set = filter_diff(patch_set, filters)
        print(
            f"Changed files: "
            f"{mc.ui.green(_what or 'INDEX')} vs "
            f"{mc.ui.yellow(_against or repo.remotes.origin.refs.HEAD.reference.name)}"
            f"{' filtered by ' + mc.ui.cyan(filters) if filters else ''}"
        )

        for patch in patch_set:
            if patch.is_added_file:
                color = mc.ui.green
            elif patch.is_removed_file:
                color = mc.ui.red
            else:
                color = mc.ui.blue
            print(f"- {color(patch.path)}")
            if diff:
                print(mc.ui.gray(textwrap.indent(str(patch), "  ")))
    finally:
        repo.close()



import logging
import os
import re
from dataclasses import dataclass, field

import git
from gito.utils import is_running_in_github_action


@dataclass
class IssueTrackerIssue:
    title: str = field(default="")
    description: str = field(default="")
    url: str = field(default="")


def extract_issue_key(branch_name: str, min_len=2, max_len=10) -> str | None:
    boundary = r'\b|_|-|/|\\'
    pattern = fr"(?:{boundary})([A-Z][A-Z0-9]{{{min_len - 1},{max_len - 1}}}-\d+)(?:{boundary})"
    match = re.search(pattern, branch_name)
    return match.group(1) if match else None


def get_branch(repo: git.Repo):
    if is_running_in_github_action():
        branch_name = os.getenv('GITHUB_HEAD_REF')
        if branch_name:
            return branch_name

        github_ref = os.getenv('GITHUB_REF', '')
        if github_ref.startswith('refs/heads/'):
            return github_ref.replace('refs/heads/', '')
    try:
        branch_name = repo.active_branch.name
        return branch_name
    except Exception as e:  # @todo: specify more precise exception
        logging.error("Could not determine the active branch name: %s", e)
        return None


def resolve_issue_key(repo: git.Repo):
    branch_name = get_branch(repo)
    if not branch_name:
        logging.error("No active branch found in the repository, cannot determine issue key.")
        return None

    if not (issue_key := extract_issue_key(branch_name)):
        logging.error(f"No issue key found in branch name: {branch_name}")
        return None
    return issue_key



import os
import sys
import io
import logging
from datetime import datetime
from pathlib import Path

import microcore as mc

from .utils import is_running_in_github_action
from .constants import HOME_ENV_PATH, EXECUTABLE, PROJECT_GITO_FOLDER
from .env import Env


def setup_logging(log_level: int = logging.INFO):
    class CustomFormatter(logging.Formatter):
        def format(self, record):
            dt = datetime.fromtimestamp(record.created).strftime("%Y-%m-%d %H:%M:%S")
            message, level_name = record.getMessage(), record.levelname
            if record.levelno == logging.WARNING:
                message = mc.ui.yellow(message)
                level_name = mc.ui.yellow(level_name)
            if record.levelno >= logging.ERROR:
                message = mc.ui.red(message)
                level_name = mc.ui.red(level_name)
            return f"{dt} {level_name}: {message}"

    handler = logging.StreamHandler()
    handler.setFormatter(CustomFormatter())
    logging.basicConfig(level=log_level, handlers=[handler])


def bootstrap(verbosity: int = 1):
    """Bootstrap the application with the environment configuration."""
    log_levels_by_verbosity = {
        0: logging.CRITICAL,
        1: logging.INFO,
        2: logging.DEBUG,
        3: logging.DEBUG,
    }
    Env.verbosity = verbosity
    Env.logging_level = log_levels_by_verbosity.get(verbosity, logging.INFO)
    setup_logging(Env.logging_level)
    logging.info(
        f"Bootstrapping Gito v{Env.gito_version}... "
        + mc.ui.gray(f"[verbosity={verbosity}]")
    )

    # cp1251 is used on Windows when redirecting output
    if sys.stdout.encoding and sys.stdout.encoding.lower() != "utf-8":
        sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')

    try:
        mc.configure(
            DOT_ENV_FILE=HOME_ENV_PATH,
            USE_LOGGING=verbosity >= 1,
            EMBEDDING_DB_TYPE=mc.EmbeddingDbType.NONE,
            PROMPT_TEMPLATES_PATH=[
                PROJECT_GITO_FOLDER,
                Path(__file__).parent / "tpl"
            ],
        )
        if verbosity > 1:
            mc.logging.LoggingConfig.STRIP_REQUEST_LINES = None
        else:
            mc.logging.LoggingConfig.STRIP_REQUEST_LINES = [300, 15]

    except mc.LLMConfigError as e:
        msg = str(e)
        if is_running_in_github_action():
            ref = os.getenv("GITHUB_WORKFLOW_REF", "")
            if ref:
                # example value: 'owner/repo/.github/workflows/ai-code-review.yml@refs/pull/1/merge'
                ref = ref.split("@")[0]
                ref = ref.split(".github/workflows/")[-1]
                ref = f" (.github/workflows/{ref})"
            msg += (
                f"\nPlease check your GitHub Action Secrets "
                f"and `env` configuration section of the corresponding workflow step{ref}."
            )
        else:
            msg += (
                f"\nPlease run '{EXECUTABLE} setup' "
                "to configure LLM API access (API keys, model, etc)."
            )
        print(mc.ui.red(msg))
        raise SystemExit(2)
    except Exception as e:
        logging.error(f"Unexpected configuration error: {e}")
        raise SystemExit(3)



import logging
import tomllib
from dataclasses import dataclass, field
from pathlib import Path

import microcore as mc
from gito.utils import detect_github_env
from microcore import ui
from git import Repo

from .constants import PROJECT_CONFIG_BUNDLED_DEFAULTS_FILE, PROJECT_CONFIG_FILE_PATH
from .pipeline import PipelineStep


@dataclass
class ProjectConfig:
    prompt: str = ""
    summary_prompt: str = ""
    answer_prompt: str = ""
    report_template_md: str = ""
    """Markdown report template"""
    report_template_cli: str = ""
    """Report template for CLI output"""
    post_process: str = ""
    retries: int = 3
    """LLM retries for one request"""
    max_code_tokens: int = 32000
    prompt_vars: dict = field(default_factory=dict)
    mention_triggers: list[str] = field(default_factory=list)
    answer_github_comments: bool = field(default=True)
    """
    Defines the keyword or mention tag that triggers bot actions
    when referenced in code review comments.
    """
    pipeline_steps: dict[str, dict | PipelineStep] = field(default_factory=dict)
    collapse_previous_code_review_comments: bool = field(default=True)
    """
    If True, previously added code review comments in the pull request
    will be collapsed automatically when a new comment is added.
    """

    def __post_init__(self):
        self.pipeline_steps = {
            k: PipelineStep(**v) if isinstance(v, dict) else v
            for k, v in self.pipeline_steps.items()
        }

    @staticmethod
    def _read_bundled_defaults() -> dict:
        with open(PROJECT_CONFIG_BUNDLED_DEFAULTS_FILE, "rb") as f:
            config = tomllib.load(f)
        return config

    @staticmethod
    def load_for_repo(repo: Repo):
        return ProjectConfig.load(Path(repo.working_tree_dir) / PROJECT_CONFIG_FILE_PATH)

    @staticmethod
    def load(config_path: str | Path | None = None) -> "ProjectConfig":
        config = ProjectConfig._read_bundled_defaults()
        github_env = detect_github_env()
        config["prompt_vars"] |= github_env | dict(github_env=github_env)

        config_path = Path(config_path or PROJECT_CONFIG_FILE_PATH)
        if config_path.exists():
            logging.info(
                f"Loading project-specific configuration from {mc.utils.file_link(config_path)}...")
            default_prompt_vars = config["prompt_vars"]
            default_pipeline_steps = config["pipeline_steps"]
            with open(config_path, "rb") as f:
                config.update(tomllib.load(f))
            # overriding prompt_vars config section will not empty default values
            config["prompt_vars"] = default_prompt_vars | config["prompt_vars"]
            # merge individual pipeline steps
            for k, v in config["pipeline_steps"].items():
                config["pipeline_steps"][k] = default_pipeline_steps.get(k, {}) | v
            # merge pipeline steps dict
            config["pipeline_steps"] = default_pipeline_steps | config["pipeline_steps"]
        else:
            logging.info(
                f"No project config found at {ui.blue(config_path)}, using defaults"
            )

        return ProjectConfig(**config)



import logging
import os
import requests

import git

from gito.issue_trackers import IssueTrackerIssue, resolve_issue_key


def fetch_issue(issue_key: str, api_key: str = None) -> IssueTrackerIssue | None:
    """
    Fetch a Linear issue using GraphQL API.
    """
    api_key = api_key or os.getenv("LINEAR_API_KEY")
    try:
        url = "https://api.linear.app/graphql"
        headers = {
            "Authorization": f"{api_key}",
            "Content-Type": "application/json"
        }

        query = """
            query Issues($teamKey: String!, $issueNumber: Float) {
                issues(filter: {team: {key: {eq: $teamKey}}, number: {eq: $issueNumber}}) {
                    nodes {
                        id
                        identifier
                        title
                        description
                        url
                    }
                }
            }
        """
        team_key, issue_number = issue_key.split("-")
        response = requests.post(
            url,
            json={
                "query": query,
                "variables": {'teamKey': team_key, 'issueNumber': int(issue_number)}
            },
            headers=headers
        )
        response.raise_for_status()
        data = response.json()

        if "errors" in data:
            logging.error(f"Linear API error: {data['errors']}")
            return None

        nodes = data.get("data", {}).get("issues", {}).get("nodes", [])
        if not nodes:
            logging.error(f"Linear issue {issue_key} not found")
            return None

        issue = nodes[0]
        return IssueTrackerIssue(
            title=issue["title"],
            description=issue.get("description") or "",
            url=issue["url"]
        )

    except requests.HTTPError as e:
        logging.error(f"Failed to fetch Linear issue {issue_key}: {e}")
        logging.error(f"Response body: {response.text}")
        return None


def fetch_associated_issue(
    repo: git.Repo,
    api_key=None,
    **kwargs
):
    """
    Pipeline step to fetch a Linear issue based on the current branch name.
    """
    api_key = api_key or os.getenv("LINEAR_API_KEY")
    if not api_key:
        logging.error("LINEAR_API_KEY environment variable is not set")
        return

    issue_key = resolve_issue_key(repo)
    return dict(
        associated_issue=fetch_issue(issue_key, api_key)
    ) if issue_key else None



import json
import logging
from dataclasses import dataclass, field, asdict
from datetime import datetime
from enum import StrEnum
from pathlib import Path

import microcore as mc
from colorama import Fore, Style, Back
from microcore.utils import file_link
import textwrap

from .constants import JSON_REPORT_FILE_NAME, HTML_TEXT_ICON, HTML_CR_COMMENT_MARKER
from .project_config import ProjectConfig
from .utils import syntax_hint, block_wrap_lr, max_line_len, remove_html_comments


@dataclass
class Issue:
    @dataclass
    class AffectedCode:
        start_line: int = field()
        end_line: int | None = field(default=None)
        file: str = field(default="")
        proposal: str = field(default="")
        affected_code: str = field(default="")

        @property
        def syntax_hint(self) -> str:
            return syntax_hint(self.file)

    id: str = field()
    title: str = field()
    details: str = field(default="")
    severity: int | None = field(default=None)
    confidence: int | None = field(default=None)
    tags: list[str] = field(default_factory=list)
    file: str = field(default="")
    affected_lines: list[AffectedCode] = field(default_factory=list)

    def __post_init__(self):
        self.affected_lines = [
            Issue.AffectedCode(**dict(file=self.file) | i)
            for i in self.affected_lines
        ]

    def github_code_link(self, github_env: dict) -> str:
        url = (
            f"https://github.com/{github_env['github_repo']}"
            f"/blob/{github_env['github_pr_sha_or_branch']}"
            f"/{self.file}"
        )
        if self.affected_lines:
            url += f"#L{self.affected_lines[0].start_line}"
            if self.affected_lines[0].end_line:
                url += f"-L{self.affected_lines[0].end_line}"
        return url


@dataclass
class Report:
    class Format(StrEnum):
        MARKDOWN = "md"
        CLI = "cli"

    issues: dict[str, list[Issue]] = field(default_factory=dict)
    summary: str = field(default="")
    number_of_processed_files: int = field(default=0)
    total_issues: int = field(init=False)
    created_at: str = field(default_factory=lambda: datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    model: str = field(default_factory=lambda: mc.config().MODEL)
    pipeline_out: dict = field(default_factory=dict)

    @property
    def plain_issues(self):
        return [
            issue
            for file, issues in self.issues.items()
            for issue in issues
        ]

    def __post_init__(self):
        issue_id: int = 0
        for file in self.issues.keys():
            self.issues[file] = [
                Issue(
                    **{
                        "id": (issue_id := issue_id + 1),
                        "file": file,
                    } | issue
                )
                for issue in self.issues[file]
            ]
        self.total_issues = issue_id

    def save(self, file_name: str = ""):
        file_name = file_name or JSON_REPORT_FILE_NAME
        with open(file_name, "w") as f:
            json.dump(asdict(self), f, indent=4)
        logging.info(f"Report saved to {mc.utils.file_link(file_name)}")

    @staticmethod
    def load(file_name: str | Path = ""):
        with open(file_name or JSON_REPORT_FILE_NAME, "r") as f:
            data = json.load(f)
        data.pop("total_issues", None)
        return Report(**data)

    def render(
        self,
        config: ProjectConfig = None,
        report_format: Format = Format.MARKDOWN,
    ) -> str:
        config = config or ProjectConfig.load()
        template = getattr(config, f"report_template_{report_format}")
        return mc.prompt(
            template,
            report=self,
            ui=mc.ui,
            Fore=Fore,
            Style=Style,
            Back=Back,
            file_link=file_link,
            textwrap=textwrap,
            block_wrap_lr=block_wrap_lr,
            max_line_len=max_line_len,
            HTML_TEXT_ICON=HTML_TEXT_ICON,
            HTML_CR_COMMENT_MARKER=HTML_CR_COMMENT_MARKER,
            remove_html_comments=remove_html_comments,
            **config.prompt_vars
        )

    def to_cli(self, report_format=Format.CLI):
        output = self.render(report_format=report_format)
        print("")
        print(output)



import contextlib
import logging
import tempfile

import microcore as mc
import typer
from git import Repo
from gito.utils import parse_refs_pair


def args_to_target(refs, what, against) -> tuple[str | None, str | None]:
    _what, _against = parse_refs_pair(refs)
    if _what:
        if what:
            raise typer.BadParameter(
                "You cannot specify both 'refs' <WHAT>..<AGAINST> and '--what'. Use one of them."
            )
    else:
        _what = what
    if _against:
        if against:
            raise typer.BadParameter(
                "You cannot specify both 'refs' <WHAT>..<AGAINST> and '--against'. Use one of them."
            )
    else:
        _against = against
    return _what, _against


def arg_refs() -> typer.Argument:
    return typer.Argument(
        default=None,
        help="Git refs to review, [what]..[against] e.g. 'HEAD..HEAD~1'"
    )


def arg_what() -> typer.Option:
    return typer.Option(None, "--what", "-w", help="Git ref to review")


def arg_filters() -> typer.Option:
    return typer.Option(
        "", "--filter", "-f", "--filters",
        help="""
            filter reviewed files by glob / fnmatch pattern(s),
            e.g. 'src/**/*.py', may be comma-separated
            """,
    )


def arg_out() -> typer.Option:
    return typer.Option(
        None,
        "--out", "-o", "--output",
        help="Output folder for the code review report"
    )


def arg_against() -> typer.Option:
    return typer.Option(
        None,
        "--against", "-vs", "--vs",
        help="Git ref to compare against"
    )


app = typer.Typer(pretty_exceptions_show_locals=False)


@contextlib.contextmanager
def get_repo_context(url: str, branch: str):
    """Context manager for handling both local and remote repositories."""
    if url:
        with tempfile.TemporaryDirectory() as temp_dir:
            logging.info(
                f"get_repo_context: "
                f"Cloning [{mc.ui.green(url)}] to {mc.utils.file_link(temp_dir)} ..."
            )
            repo = Repo.clone_from(url, branch=branch, to_path=temp_dir)
            try:
                yield repo, temp_dir
            finally:
                repo.close()
    else:
        logging.info("get_repo_context: Using local repo...")
        repo = Repo(".")
        try:
            yield repo, "."
        finally:
            repo.close()
